;nyquist plug-in
;version 3
;type analyze
;categories "http://lv2plug.in/ns/lv2core#AnalyserPlugin"
;name "Advanced Sound Finder..."
;action "Analyzing..."
;info "by Steve Daulton. (www.easyspacepro.com)\nReleased under GPL v2."


;; AdvancedSoundFinder.ny by Steve Daulton May 2012.
;; Released under terms of the GNU General Public License version 2:
;; http://www.gnu.org/licenses/old-licenses/gpl-2.0.html .


;control thresh "Threshold for silence" real "dB" -30 -60 0
;control peak-or-rms "Threshold detection method" choice "Peak Level,RMS Level" 0
;control sgap "Allow gaps in sound less than" real "seconds" 1 0.1 5
;control minsnd "Disregard isolated sounds less than" real "seconds" 10 0 60
;control minlabel "Minimum label length (optional)" string "min:seconds" "0:00.000" ""
;control tbefore "Label starting point (before start of sound)" real "seconds" 0.1 0 1
;control tafter "Label ending point (after end of sound)" real "seconds" 0.1 0 1
;control snum "Begin numbering from" string "" "1" ""
;control digits "Minimum number of digits" choice "1,2,3,4" 1 
;control text "Label text (optional)" string  "Track_"
;control opt "Label options" choice "Number only,Text only,Number & Text,Text & Number" 3

;; Note: label position accuracy with rms level detection will not be better than +/- 10 ms.
;; "Minimum label length (optional)" could be renamed as "Minimum Sound Group Duration" or "Group Sounds (optional)"


; thresh        ; silence threshold [-96dB to 0 dB]
; peak-or-rms   ; peak or rms [Peak/RMS] rms window size=20 ms 
; sgap          ; allowed gap in sound [> 0.0 seconds]
; minsnd        ; minimum sound duration to detect [> 0.0 seconds]
; minlabel      ; minimum label length [text input min:sec > 0.0 seconds]
; tbefore       ; label start before sound [> 0.0 seconds]
; tafter        ; label end after sound [> 0.0 seconds]
; snum          ; first number [integer]
; digits        ; number of digits [1 to 4]
; text          ; optional label text [string]
; opt           ; Label options [0=number, 1=text, 2=number+text, 3=text+number]

;;; globals
(setq err "")                   ;initialise error message
(setq thresh (db-to-linear (min (max -96 thresh) 0))) ;silence threshold
(setq sgap (max 0.01 sgap))     ;allowed gap in sound
(setq minsnd (max 0 minsnd))    ;minimum sound duration to detect
(setq tbefore (max 0 tbefore))  ;label start before sound
(setq tafter (max 0 tafter))    ;label end after sound
(setq digits (1+ digits))       ;number of digits
(setq LowRate 0)                ;initialise low sample rate
(setf SndList ())               ;List of Sound Start points
(setf SilList ())               ;List of Sound End points
(setf LabelList ())             ;Label List for output


;;; Convert a string into a list
(defun string-to-list (string)
  (read (make-string-input-stream (format nil "(~a)" string))))


;;; Stereo channels to mono. 
(defun mono-s (s-in) 
  (if (arrayp s-in)
    (s-max (aref s-in 0)(aref s-in 1)) 
    s-in))


;;; Resamples sound to low sample rate. peak-or-rms: 0=peak 1=rms
(defun lo-srate (s-in)
  ;; low sample rate ~ 400 Hz
  (let ((step (round (/ *sound-srate* 400.0)))
        (sound (mono-s s-in)))
    (if (= peak-or-rms 0)
      (snd-avg sound step step OP-PEAK)
      (rms (mono-s s) 50))))


;;; Substitute character in string
(defun str-subst-char (txt oldchar newchar)
  (do ((ch 0 (1+ ch))(newstring ""))
      ((= ch (length txt)) newstring)
    (let ((next (char txt ch)))
      (setf newstring 
        (if (char= next oldchar)
            (strcat newstring (string newchar))
            (strcat newstring (string next)))))))

;;; Format number to appropriate digits and return a string
(defun numstring ()
  (setq nstr (format nil "~a" snum))
  (do ((n (length nstr) (1+ n)))
      ((>= n digits) nstr)
    (setq nstr (strcat "0" nstr))))


;;; Make label string 0=number, 1=text, 2=number+text, 3=text+number
(defun make-label-string ()
  (case opt
    (0 (setq lstring (format nil "~a" (numstring)))
       (setq snum (1+ snum)))
    (1 (setq lstring text))
    (2 (setq lstring (format nil "~a~a" (numstring) text))
       (setq snum (1+ snum)))
    (T (setq lstring (format nil "~a~a" text (numstring)))
       (setq snum (1+ snum))))
  lstring)


;;; Remove short silences and add to LabelList
(defun AllowSilence (snd sil)
  (let ((snd (reverse snd))                 ;put into time order
        (sil (reverse sil))
        (sgap (* sgap LowRate))             ;convert to samples
        (NextL 0)                           ;sound label counter
        (num (1- (length snd))))            ;nth sample number
    
    (dotimes (SndCount num)
      (setq TempStart (nth NextL snd))
      (if (> (- (nth (1+ SndCount) snd)(nth SndCount sil)) sgap)
        (progn
          (setf LabelList (cons (list TempStart (nth SndCount sil)) LabelList))
          (setq NextL (1+ SndCount)))))
    (setf LabelList 
      (cons (list (nth NextL snd)(nth num sil)) LabelList))))


;;; Remove short sounds
(defun IgnoreSound ()
  (let ((minsnd (* minsnd LowRate))   ;convert to samples
        (TempLabels ())                     ;Temp label list
        (num (length LabelList)))           ;nth label number
    (dotimes (SndCount num)
      (setf tmplab (nth SndCount LabelList))
      (if (>= (- (second tmplab)(first tmplab)) minsnd)
        (setf TempLabels (cons tmplab TempLabels))))
    (setf LabelList TempLabels)))


;;; Don't split shorter than (seconds)
(defun SplitLen ()
  (let ((minlabel (* minlabel LowRate))   ;convert to samples
        (TempLabels ())                         ;Temp label list
        (num (length LabelList))               ;nth label number
        (addlabel 1))
    ;(setq tmplab ())
    (dotimes (SndCount num)
      (if (= addlabel 0)
        (setf tmplab (list (first tmplab)(second (nth SndCount LabelList))))
        (setq tmplab (nth SndCount LabelList)))
      (if (>= (- (second tmplab)(first tmplab)) minlabel)
        (progn
          (setf TempLabels (cons tmplab TempLabels))
          (setq addlabel 1))
        (setq addlabel 0)))
    (if (= addlabel 0)
      (setf TempLabels (cons tmplab TempLabels)))  ;add last label
    (setf LabelList (reverse TempLabels))))


;;; Put data into label format
(defun MakeLabelTrack (LabelList)
  ; low sample rate compensation
  (setq sr-comp (if (= peak-or-rms 0) 0.0025 0.02))
  ;;; Adjust label start
  (defun label-start (start-time)
    (let ((start-time (- start-time tbefore sr-comp)))
      (max 0 start-time)))
  ;;; Adjust label end
  (defun label-end (end-time)
    (- (+ end-time tafter) sr-comp))
  ;;; Format Label
  (defun FormatLabel (Label)
    (list 
      (label-start (/ (first Label) LowRate))
      (label-end (/ (second Label) LowRate))
      (make-label-string)))    
  ;; step through LabelList
  (dotimes (val (length LabelList))
    (setf (nth val LabelList) (FormatLabel (nth val LabelList)))))
        

;;; Apply sort filters and convert to label track format
(defun sortlist (snd sil)
  (prog ()
        (allowsilence snd sil)
        (IgnoreSound)
        (when (= (length LabelList) 0)
          (setq err (strcat err "Try setting 'Disregard isolated sounds' to a smaller value.\n"))
          (go end))
        (SplitLen)
        (MakeLabelTrack LabelList)
   end  (if (> (length err) 0)
          (return (format nil "Warning.~%No labels produced.~%~a" err)) ;there should only be warnings and not errors here.
          (return LabelList))))   ;return label track


;;; Build list of sound start and end times
(defun BuildList ()
  (let ((SndFound 0)
        (SilFound 0)
        (s1 (lo-srate s)))
    (setq s nil)                    ;free s to save memory
    (setq LowRate (snd-srate s1))   ;compute sampe rate
    (setq s1len
      (do ((sample 1 (1+ sample))
           (sVal (snd-fetch s1) (setq sVal (snd-fetch s1))))
          ((not sVal) sample) ;exit when out of samples, return number of samples
      
      (if (> sVal Thresh)           ;level above Thresh
        (progn
          (setq SilFound 0)         ;reset SilFound
          (if (= SndFound 0)        ;level crossed Thresh
            (progn
              (setf SndList (cons sample SndList))
              (setq SndFound 1))))
        (progn                      ;level below Thresh
          (setq SndFound 0)         ;reset SndFound
          (if (= SilFound 0)
            (progn
              (if (> (length SndList) 0) ;sound list must start first
                (setf SilList (cons sample SilList)))
              (setq SilFound 1)))))))
    (if (= SilFound 0)                        ;if end of last sound not found
      (setf SilList (cons s1len SilList)))))   ;add end of selection
      

;;; convert minimum label length to time (seconds)
(setf minlabel (str-subst-char minlabel #\: #\space)) ;replace colons with spaces
(setf minlabel (str-subst-char minlabel #\, #\.))     ;replace commas with dots
(setf minlabel (string-to-list minlabel))
;; check format of minlabel
(let ((errcount 0))
  (dotimes (el (length minlabel))
    (if (not (numberp (nth el minlabel)))
      (setq errcount (1+ errcount))))
  (when (> errcount 0)
        (setq err (strcat err (format nil "Invalid Minimum label length.~%")))
        (setf minlabel (list 0))))
;; calculate min label length
(case (length minlabel)
      (0 (setq minlabel 0))
      (1 (setq minlabel (max 0 (first minlabel))))
      (2 (setq minlabel (max 0 (+ (* 60 (first minlabel))(second minlabel))))))

;;; first label number
(setq snum (first (string-to-list snum)))
(setq snum
  (if snum
    (if (numberp snum)
        (truncate snum) ;ensure integer
        (progn
          (setq err (strcat err (format nil "'Begin numbering from' must be a number.~%")))
          1))           ;error and default to 1
        
    1))                 ;else default to 1








(if (> (length err) 0)
  (format nil "Error.~%~a" err)
  (progn 
    (BuildList)   ;build label lists
    (if (null SndList)
      (format nil "Warning.~%No sounds found.~%Try lowering the Sound Threshold level.")
      (SortList SndList SilList)))) ;returns label track or warning message.

